(ns status-im.node.core
  (:require [re-frame.core :as re-frame]
            [status-im.constants :as constants]
            [status-im.ethereum.json-rpc :as json-rpc]
            [status-im.native-module.core :as status]
            [status-im.utils.config :as config]
            [status-im.utils.fx :as fx]
            [status-im.utils.platform :as utils.platform]
            [status-im.utils.types :as types]
            [taoensso.timbre :as log]
            [status-im.ethereum.ens :as ens])
  (:require-macros [status-im.utils.slurp :refer [slurp]]))

(defn- add-custom-bootnodes [config network all-bootnodes]
  (let [bootnodes (as-> all-bootnodes $
                    (get $ network)
                    (vals $)
                    (map :address $))]
    (if (seq bootnodes)
      (assoc-in config [:ClusterConfig :BootNodes] bootnodes)
      config)))

(defn- add-log-level [config log-level]
  (if (empty? log-level)
    (assoc config
           :LogLevel "ERROR"
           :LogEnabled false)
    (assoc config
           :LogLevel log-level
           :LogEnabled true)))

(defn get-network-genesis-hash-prefix
  "returns the hex representation of the first 8 bytes of
  a network's genesis hash"
  [network]
  (case network
    1 "d4e56740f876aef8"
    3 "41941023680923e0"
    4 "6341fd3daf94b748"
    nil))

(defn get-les-topic
  "returns discovery v5 topic derived from genesis of the provided network"
  [network]
  (let [les-discovery-identifier "LES2@"
        hash-prefix (get-network-genesis-hash-prefix network)]
    (when hash-prefix
      (str les-discovery-identifier hash-prefix))))

(defn get-topics
  [network]
  (let [les-topic (get-les-topic network)]
    (cond-> {"whisper" {:Min 2, :Max 2}}
      les-topic (assoc les-topic {:Min 2, :Max 2}))))

(defn- get-base-node-config [config]
  (let [initial-props @(re-frame/subscribe [:initial-props])
        status-node-port (get initial-props :STATUS_NODE_PORT)]
    (cond-> (assoc config
                   :Name "StatusIM")
      config/dev-build?
      (assoc :ListenAddr ":30304"
             :DataDir (str (:DataDir config) "_dev"))
      status-node-port
      (assoc :ListenAddr (str ":" status-node-port)))))

(defn- pick-nodes
  "Picks `limit` different nodes randomly from the list of nodes
  if there is more than `limit` nodes in the list, otherwise return the list
  of nodes"
  [limit nodes]
  (take limit (shuffle nodes)))

(def default-fleets (slurp "resources/config/fleets.json"))

(defn fleets [{:keys [custom-fleets]}]
  (as-> [(default-fleets)] $
    (mapv #(:fleets (types/json->clj %)) $)
    (conj $ custom-fleets)
    (reduce merge $)))

(defn current-fleet-key [db]
  (keyword (get-in db [:multiaccount :fleet]
                   config/fleet)))

(defn get-current-fleet
  [db]
  (get (fleets db)
       (current-fleet-key db)))

(defn- get-multiaccount-node-config
  [{:keys [multiaccount :networks/networks :networks/current-network]
    :as db}]
  (let [current-fleet-key (current-fleet-key db)
        current-fleet (get-current-fleet db)
        rendezvous-nodes (pick-nodes 3 (vals (:rendezvous current-fleet)))
        {:keys [installation-id log-level
                waku-enabled waku-bloom-filter-mode
                custom-bootnodes custom-bootnodes-enabled?]} multiaccount
        use-custom-bootnodes (get custom-bootnodes-enabled? current-network)]
    (cond-> (get-in networks [current-network :config])
      :always
      (get-base-node-config)

      current-fleet
      (assoc :NoDiscovery   false
             :Rendezvous    (not (empty? rendezvous-nodes))
             :ClusterConfig {:Enabled true
                             :Fleet              (name current-fleet-key)
                             :BootNodes
                             (pick-nodes 4 (vals (:boot current-fleet)))
                             :TrustedMailServers
                             (pick-nodes 6 (vals (:mail current-fleet)))
                             :StaticNodes
                             (into (pick-nodes 2
                                               (vals (:whisper current-fleet)))
                                   (vals (:static current-fleet)))
                             :RendezvousNodes    rendezvous-nodes})

      :always
      (assoc :WalletConfig {:Enabled true}
             :BrowsersConfig {:Enabled true}
             :PermissionsConfig {:Enabled true}
             :MailserversConfig {:Enabled true}
             :EnableNTPSync true
             (if waku-enabled :WakuConfig :WhisperConfig)
             {:Enabled true
              :BloomFilterMode waku-bloom-filter-mode
              :LightClient true
              :MinimumPoW 0.001}
             :ShhextConfig
             {:BackupDisabledDataDir      (utils.platform/no-backup-directory)
              :InstallationID             installation-id
              :MaxMessageDeliveryAttempts config/max-message-delivery-attempts
              :MailServerConfirmations    config/mailserver-confirmations-enabled?
              :VerifyTransactionURL       constants/mainnet-rpc-url
              :VerifyENSURL               constants/mainnet-rpc-url
              :VerifyENSContractAddress   (:mainnet ens/ens-registries)
              :VerifyTransactionChainID   1
              :DataSyncEnabled            true
              :PFSEnabled                 true}
             :RequireTopics (get-topics current-network)
             :StatusAccountsConfig {:Enabled true})

      (and
       config/bootnodes-settings-enabled?
       use-custom-bootnodes)
      (add-custom-bootnodes current-network custom-bootnodes)

      :always
      (add-log-level log-level))))

(defn get-new-config
  [db]
  (types/clj->json (get-multiaccount-node-config db)))

(fx/defn save-new-config
  "Saves a new status-go config for the current account
   This RPC method is the only way to change the node config of an account.
   NOTE: it is better used indirectly through `prepare-new-config`,
    which will take care of building up the proper config based on settings in
app-db"
  {:events [::save-new-config]}
  [{:keys [db]} config {:keys [on-success]}]
  {::json-rpc/call [{:method "settings_saveSetting"
                     :params [:node-config config]
                     :on-success on-success}]})

(fx/defn prepare-new-config
  "Use this function to apply settings to the current account node config"
  [{:keys [db]} {:keys [on-success]}]
  {::prepare-new-config [(get-new-config db)
                         #(re-frame/dispatch
                           [::save-new-config % {:on-success on-success}])]})

(re-frame/reg-fx
 ::prepare-new-config
 (fn [[config callback]]
   (status/prepare-dir-and-update-config config callback)))
